use bevy::{
    input::mouse::MouseMotion,
    pbr::wireframe::{
        WireframeConfig,
        WireframePlugin,
    },
    prelude::*,
    render::{
        mesh::Indices,
        render_resource::PrimitiveTopology,
        settings::{
            WgpuFeatures,
            WgpuSettings,
        },
        RenderPlugin,
    },
    window::CursorGrabMode,
};
use sf_mesher::{
    chunk::Chunk,
    simple::simple_mesh,
    voxel::{
        Visibility,
        Voxel,
    },
};

#[derive(Copy, Clone, Default, PartialEq, Eq)]
pub enum MyVoxel {
    #[default]
    Empty,
    Opaque(u16),
    Transparent(u16),
}

impl Voxel for MyVoxel {
    fn visibility(&self) -> Visibility {
        match self {
            Self::Empty => Visibility::Empty,
            Self::Opaque(_) => Visibility::Opaque,
            Self::Transparent(_) => Visibility::Transparent,
        }
    }
}

pub const CHUNK_SIZE: usize = 32;

pub struct MyChunk {
    voxels: Box<[MyVoxel; CHUNK_SIZE * CHUNK_SIZE * CHUNK_SIZE]>,
}

impl Default for MyChunk {
    fn default() -> Self {
        Self {
            voxels: Box::new([MyVoxel::default(); CHUNK_SIZE * CHUNK_SIZE * CHUNK_SIZE]),
        }
    }
}

impl Chunk for MyChunk {
    type Output = MyVoxel;

    const X: usize = CHUNK_SIZE;
    const Y: usize = CHUNK_SIZE;
    const Z: usize = CHUNK_SIZE;

    fn get(&self, x: usize, y: usize, z: usize) -> Self::Output {
        self.voxels[Self::linearize(x, y, z)]
    }
}

#[derive(Component, Default)]
struct PlayerCamera;

#[derive(Component, Default)]
struct Player;

#[derive(Resource)]
struct Sensitivity(f32);

fn setup_wireframe(mut wireframe_config: ResMut<WireframeConfig>) {
    wireframe_config.global = true;
}

fn setup_player(mut commands: Commands) {
    commands
        .spawn(())
        .insert(Name::from("Player"))
        .insert(Player)
        .insert(SpatialBundle::default())
        .with_children(|commands| {
            commands.spawn((PlayerCamera, Camera3dBundle::default()));
        });
}

fn setup_lights(mut commands: Commands) {
    commands.spawn(DirectionalLightBundle {
        directional_light: DirectionalLight {
            illuminance: 100000.0,
            shadows_enabled: true,
            ..default()
        },

        transform: Transform {
            translation: Vec3::new(0.0, 2.0, 0.0),
            rotation: Quat::from_rotation_x(-std::f32::consts::FRAC_PI_4),
            ..default()
        },

        ..default()
    });
}

fn setup_mouse_lock(mut windows: Query<&mut Window>) {
    if let Ok(mut window) = windows.get_single_mut() {
        window.cursor.grab_mode = CursorGrabMode::Confined;
        window.cursor.visible = false;
    }
}

fn mouse_look(
    time: Res<Time>,
    mut input: EventReader<MouseMotion>,
    sensitivity: Res<Sensitivity>,
    mut camera: Query<&mut Transform, With<PlayerCamera>>,
    mut player: Query<&mut Transform, (With<Player>, Without<PlayerCamera>)>,
) {
    let mut camera_tf = camera.single_mut();
    let mut player_tf = player.single_mut();

    let dt = time.delta_seconds();
    let sens = sensitivity.0;

    for motion in input.iter() {
        let rot = camera_tf.rotation;
        camera_tf.rotate(Quat::from_scaled_axis(
            rot * Vec3::X * -motion.delta.y * dt * sens,
        ));

        let rot = player_tf.rotation;
        player_tf.rotate(Quat::from_scaled_axis(
            rot * Vec3::Y * -motion.delta.x * dt * sens,
        ));
    }
}

fn mouse_lock(
    mouse: Res<Input<MouseButton>>,
    keyboard: Res<Input<KeyCode>>,
    mut windows: Query<&mut Window>,
) {
    if let Ok(mut window) = windows.get_single_mut() {
        if mouse.just_pressed(MouseButton::Left) {
            window.cursor.grab_mode = CursorGrabMode::Confined;
            window.cursor.visible = false;
        }

        if keyboard.just_pressed(KeyCode::Escape) {
            window.cursor.grab_mode = CursorGrabMode::None;
            window.cursor.visible = true;
        }
    }
}

fn player_move(
    keys: Res<Input<KeyCode>>,
    time: Res<Time>,
    windows: Query<&Window>,
    mut query: Query<&mut Transform, With<Player>>,
) {
    if let Ok(window) = windows.get_single() {
        for mut transform in query.iter_mut() {
            let mut velocity = Vec3::ZERO;
            let local_z = transform.local_z();
            let forward = -Vec3::new(local_z.x, 0., local_z.z);
            let right = Vec3::new(local_z.z, 0., -local_z.x);

            for key in keys.get_pressed() {
                match window.cursor.grab_mode {
                    CursorGrabMode::None => (),
                    _ => match key {
                        KeyCode::W => velocity += forward,
                        KeyCode::S => velocity -= forward,
                        KeyCode::A => velocity -= right,
                        KeyCode::D => velocity += right,
                        KeyCode::Space => velocity += Vec3::Y,
                        KeyCode::LShift => velocity -= Vec3::Y,
                        _ => (),
                    },
                }
            }

            velocity = velocity.normalize_or_zero();

            transform.translation += velocity * time.delta_seconds() * 5.0
        }
    } else {
        warn!("Primary window not found for `player_move`!");
    }
}

fn setup_chunk(
    mut commands: Commands,
    mut meshes: ResMut<Assets<Mesh>>,
    mut materials: ResMut<Assets<StandardMaterial>>,
) {
    let mut chunk = MyChunk::default();

    for i in 0..MyChunk::size() {
        let (x, y, z) = MyChunk::delinearize(i);

        let voxel = if ((x * x + y * y + z * z) as f32).sqrt() < 8.0 {
            if y > 4 {
                MyVoxel::Transparent(1)
            } else {
                MyVoxel::Opaque(1)
            }
        } else {
            MyVoxel::Empty
        };

        chunk.voxels[i] = voxel;
    }

    let result = simple_mesh(&chunk);

    let mut positions = Vec::new();
    let mut indices = Vec::new();
    let mut normals = Vec::new();
    let mut uvs = Vec::new();

    for face in result.iter() {
        positions.extend_from_slice(&face.positions(1.0));
        indices.extend_from_slice(&face.indices(positions.len() as u32));
        normals.extend_from_slice(&face.normals());
        uvs.extend_from_slice(&face.uvs(false, true));
    }

    let mut mesh = Mesh::new(PrimitiveTopology::TriangleList);

    mesh.set_indices(Some(Indices::U32(indices)));

    mesh.insert_attribute(Mesh::ATTRIBUTE_POSITION, positions);
    mesh.insert_attribute(Mesh::ATTRIBUTE_NORMAL, normals);
    mesh.insert_attribute(Mesh::ATTRIBUTE_UV_0, uvs);

    let mesh = meshes.add(mesh);
    let material = materials.add(Color::WHITE.into());

    commands.spawn(PbrBundle {
        mesh,
        material,
        ..default()
    });
}

#[rustfmt::skip]
fn main() {
    App::new()
        .add_plugins(
            DefaultPlugins.set(RenderPlugin { 
                wgpu_settings: WgpuSettings { 
                    features: WgpuFeatures::POLYGON_MODE_LINE,
                    ..default() 
                },
            })
        )
        .add_plugin(WireframePlugin)

        .insert_resource(Sensitivity(0.15))

        .add_startup_system(setup_wireframe)
        .add_startup_system(setup_player)
        .add_startup_system(setup_chunk)
        .add_startup_system(setup_lights)
        .add_startup_system(setup_mouse_lock)

        .add_system(mouse_look)
        .add_system(mouse_lock)
        .add_system(player_move)
        
        .run();
}
