use std::ops::Deref;

use crate::{
    ao::face_aos,
    Chunk,
    Voxel,
};

#[derive(Copy, Clone, Debug)]
pub struct Quad {
    pub voxel: [usize; 3],
}

#[derive(Default)]
pub struct QuadGroups {
    pub groups: [Vec<Quad>; 6],
}

impl QuadGroups {
    #[inline]
    pub fn clear(&mut self) {
        self.groups.iter_mut().for_each(|g| g.clear());
    }

    #[inline]
    pub fn iter(&self) -> impl Iterator<Item = Face> {
        self.groups
            .iter()
            .enumerate()
            .flat_map(|(index, quads)| quads.iter().map(move |quad| (index, quad)))
            .map(|(index, quad)| Face {
                side: index.into(),
                quad,
            })
    }

    #[inline]
    pub fn iter_with_ao<'a, C, V>(&'a self, chunk: &'a C) -> impl Iterator<Item = FaceWithAO<'a>>
    where
        C: Chunk<Output = V>,
        V: Voxel,
    {
        self.iter().map(|face| FaceWithAO::new(face, chunk))
    }
}

#[derive(Copy, Clone, Debug, PartialEq, Eq)]
pub enum Axis {
    X,
    Y,
    Z,
}

#[derive(Copy, Clone, Debug, PartialEq, Eq)]
pub struct Side {
    pub axis: Axis,
    pub positive: bool,
}

impl Side {
    pub const X_NEG: Self = Self::new(Axis::X, false);
    pub const X_POS: Self = Self::new(Axis::X, true);
    pub const Y_NEG: Self = Self::new(Axis::Y, false);
    pub const Y_POS: Self = Self::new(Axis::Y, true);
    pub const Z_NEG: Self = Self::new(Axis::Z, false);
    pub const Z_POS: Self = Self::new(Axis::Z, true);

    #[inline]
    pub const fn new(axis: Axis, positive: bool) -> Self {
        Self { axis, positive }
    }

    #[rustfmt::skip]
    #[inline]
    pub fn normal(&self) -> [f32; 3] {
        match *self {
            Self::X_NEG => [-1.0, 0.0,  0.0],
            Self::X_POS  => [1.0,  0.0,  0.0],
            Self::Y_NEG => [0.0, -1.0,  0.0],
            Self::Y_POS   => [0.0,  1.0,  0.0],
            Self::Z_NEG => [0.0,  0.0, -1.0],
            Self::Z_POS   => [0.0,  0.0,  1.0],
        }
    }

    #[inline]
    pub fn normals(&self) -> [[f32; 3]; 4] {
        [self.normal(), self.normal(), self.normal(), self.normal()]
    }
}

#[rustfmt::skip]
impl From<usize> for Side {
    #[inline]
    fn from(value: usize) -> Self {
        match value {
            0 => Self::X_NEG,
            1 => Self::X_POS,
            2 => Self::Y_NEG,
            3 => Self::Y_POS,
            4 => Self::Z_NEG,
            5 => Self::Z_POS,
            _ => unreachable!()
        }
    }
}

pub struct Face<'a> {
    side: Side,
    quad: &'a Quad,
}

impl<'a> Face<'a> {
    #[inline]
    pub fn indices(&self, start: u32) -> [u32; 6] {
        [start, start + 2, start + 1, start + 1, start + 2, start + 3]
    }

    #[inline]
    pub fn positions(&self, voxel_size: f32) -> [[f32; 3]; 4] {
        #[rustfmt::skip]
        let positions = match self.side {
            Side::X_NEG => [[0.0, 0.0, 1.0], [0.0, 0.0, 0.0], [0.0, 1.0, 1.0], [0.0, 1.0, 0.0]],
            Side::X_POS => [[1.0, 0.0, 0.0], [1.0, 0.0, 1.0], [1.0, 1.0, 0.0], [1.0, 1.0, 1.0]],
            Side::Y_NEG => [[0.0, 0.0, 1.0], [1.0, 0.0, 1.0], [0.0, 0.0, 0.0], [1.0, 0.0, 0.0]],
            Side::Y_POS => [[0.0, 1.0, 1.0], [0.0, 1.0, 0.0], [1.0, 1.0, 1.0], [1.0, 1.0, 0.0]],
            Side::Z_NEG => [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [1.0, 1.0, 0.0]],
            Side::Z_POS => [[1.0, 0.0, 1.0], [0.0, 0.0, 1.0], [1.0, 1.0, 1.0], [0.0, 1.0, 1.0]],
        };

        let (x, y, z) = (
            (self.quad.voxel[0] - 1) as f32,
            (self.quad.voxel[1] - 1) as f32,
            (self.quad.voxel[2] - 1) as f32,
        );

        [
            [
                x * voxel_size + positions[0][0] * voxel_size,
                y * voxel_size + positions[0][1] * voxel_size,
                z * voxel_size + positions[0][2] * voxel_size,
            ],
            [
                x * voxel_size + positions[1][0] * voxel_size,
                y * voxel_size + positions[1][1] * voxel_size,
                z * voxel_size + positions[1][2] * voxel_size,
            ],
            [
                x * voxel_size + positions[2][0] * voxel_size,
                y * voxel_size + positions[2][1] * voxel_size,
                z * voxel_size + positions[2][2] * voxel_size,
            ],
            [
                x * voxel_size + positions[3][0] * voxel_size,
                y * voxel_size + positions[3][1] * voxel_size,
                z * voxel_size + positions[3][2] * voxel_size,
            ],
        ]
    }

    #[inline]
    pub fn normals(&self) -> [[f32; 3]; 4] {
        self.side.normals()
    }

    #[rustfmt::skip]
    #[inline]
    pub fn uvs(&self, flip_u: bool, flip_v: bool) -> [[f32; 2]; 4] {
        match (flip_u, flip_v) {
            (true, true)   => [[1.0, 1.0], [0.0, 1.0], [1.0, 0.0], [0.0, 0.0]],
            (true, false)  => [[1.0, 0.0], [0.0, 0.0], [1.0, 1.0], [0.0, 1.0]],
            (false, true)  => [[0.0, 1.0], [1.0, 1.0], [0.0, 0.0], [1.0, 0.0]],
            (false, false) => [[0.0, 0.0], [1.0, 0.0], [0.0, 1.0], [1.0, 1.0]],
        }
    }

    #[inline]
    pub fn voxel(&self) -> [usize; 3] {
        self.quad.voxel
    }

    #[inline]
    pub fn side(&self) -> Side {
        self.side
    }
}

pub struct FaceWithAO<'a> {
    face: Face<'a>,
    aos: [u32; 4],
}

impl<'a> Deref for FaceWithAO<'a> {
    type Target = Face<'a>;

    #[inline]
    fn deref(&self) -> &Self::Target {
        &self.face
    }
}

impl<'a> FaceWithAO<'a> {
    #[inline]
    pub fn new<C, V>(face: Face<'a>, chunk: &C) -> Self
    where
        C: Chunk<Output = V>,
        V: Voxel,
    {
        let aos = face_aos(&face, chunk);
        Self { face, aos }
    }

    #[inline]
    pub fn indices(&self, start: u32) -> [u32; 6] {
        let aos = self.ao_values();

        if (aos[1] + aos[2]) > (aos[0] + aos[3]) {
            [start, start + 2, start + 1, start + 1, start + 2, start + 3]
        } else {
            [start, start + 3, start + 1, start, start + 2, start + 3]
        }
    }

    #[inline]
    pub fn ao_values(&self) -> [u32; 4] {
        self.aos
    }
}
